(**************************************************************************)
(*                                                                        *)
(*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       *)
(*  Kim Nguyen, Laurent Sartre                                            *)
(*                                                                        *)
(*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      *)
(*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 *)
(*                                                                        *)
(*  https://www.informatique-mpi.fr/                                      *)
(*                                                                        *)
(**************************************************************************)

module Make(G: sig

  type player

  type state

  val player: state -> player
    (* À qui de jouer ? *)

  val moves: state -> state list
    (* Renvoie la liste des coups possibles.
       Un état est terminal si la liste est vide *)

  val outcome: state -> player option
    (* Renvoie `Some p` si le joueur `p` gagne et `None` pour un match nul.
       S'applique uniquement à un état terminal. *)

  val eval: player -> state -> int
    (* `eval p s` évalue l'état `s` du point de vue du `p`,
       entre -10000 (défaite) et +10000 (victoire) *)

end) = struct

  open G

  (* Calcul des attracteurs

   De manière générale, attr(p,i) est l'ensemble des états où le
   joueur p possède une stratégie gagnante en au plus i coups. *)
  module S =
    Set.Make(struct type t = state let compare = Stdlib.compare end)

  let winning all p =
    let rec compute att i =
      (* attention à ne pas utiliser directement List.for_all *)
      let for_all p = function [] -> false | l -> List.for_all p l in
      let add s acc =
        let test = if player s = p then List.exists else for_all in
        if test (fun s' -> S.mem s' att) (moves s) then S.add s acc else acc
      in
      let att' = S.fold add all att in
      if S.cardinal att' > S.cardinal att then compute att' (i+1) else att
    in
    let start =
      S.filter
        (fun s -> moves s = [] && player s <> p && outcome s = Some p) all in
    compute start 0

  (* Min-max.
     On maximise pour le joueur p. *)

  (* On calcule jusqu'au bout. *)

  let rec minmax (p: player) (s: state) : int =
    match moves s with
    | [] ->
        (match outcome s with
         | Some p' -> if p' = p then +1 else -1
         | None    -> 0)
    | s' :: l ->
        List.fold_left (if player s = p then max else min)
          (minmax p s') (List.map (minmax p) l)

  let rec minmax p s =
    match moves s with
    | [] ->
        (match outcome s with
         | Some p' -> if p' = p then +1 else -1
         | None    -> 0)
    | l ->
        let f, m = if player s = p then max, min_int else min, max_int in
        List.fold_left f m (List.map (fun s' -> minmax p s') l)

  (* On calcule jusqu'à la profondeur d seulement. *)

  let rec minmax (p: player) (d: int) (s: state) : int =
    match moves s with
    | [] ->
        (match outcome s with
         | Some p' -> if p' = p then +10000 else -10000
         | None    -> 0)
    | _ when d = 0 ->
        eval p s
    | s' :: l ->
        List.fold_left
          (if player s = p then max else min)
          (minmax p (d-1) s')
          (List.map (fun s' -> minmax p (d-1) s') l)

  (* Alpha-beta.
     On maximise pour le joueur p.
     On calcule jusqu'à la profondeur d. *)

  let rec alphabeta (p: player) (d: int) (a: int) (b: int) (s: state) : int =
    match moves s with
    | [] ->
        (match outcome s with
         | Some p' -> if p' = p then +10000 else -10000
         | None    -> 0)
    | _ when d = 0 ->
        eval p s
    | l ->
        let rec loop v a b = function
          | [] -> v
          | s' :: l ->
              let e = alphabeta p (d-1) a b s' in
              if player s = p then (* max *)
                let v = max v e in
                if v >= b then v else loop v (max a v) b l
              else (* min *)
                let v = min v e in
                if v <= a then v else loop v a (min b v) l
        in
        loop (if player s = p then min_int else max_int) a b l

  let alphabeta p d s =
    alphabeta p (-10000) (+10000) d s

end
